#include <u.h>
#include "console.h"

enum {
	Width = 80*2,
	Height = 25,

	/* high background, low foreground */
	Attr = (Black<<4)|Grey,
};

extern	void	bcopy(const void*, void*, int);

static __inline unsigned char
inb(int port)
{
	unsigned char data;
	__asm __volatile("inb %w1,%0" : "=a" (data) : "d" (port));
	return data;
}

static __inline void
outb(int port, unsigned char data)
{
	__asm __volatile("outb %0,%w1" : : "a" (data), "d" (port));
}

static	int	cgapos;
static	unsigned char	*cgamem = (char*)0xB8000;

static unsigned char
cgaregr(int index)
{
	outb(0x3D4, index);
	return inb(0x3D4+1)&0xFF;
}

static void
cgaregw(int index, int data)
{
	outb(0x3D4, index);
	outb(0x3D4+1, data);
}

static void
movecursor(void)
{
	cgaregw(0x0E, (cgapos/2>>8) & 0xFF);
	cgaregw(0x0F, cgapos/2 & 0xFF);
	cgamem[cgapos+1] = Attr;
}

void
cons_init(void)
{
	cgapos = cgaregr(0x0E)<<8;
	cgapos |= cgaregr(0x0F);
	cgapos *= 2;
}

void
cons_size(int *w, int *h)
{
	*w = 80;
	*h = 25;
	return;
}

void
cons_setcursor(int col, int row)
{
	if(col < 0 || col >= 80)
		return;
	if(row < 0 || row >= 25)
		return;
	cgapos = Width*row+2*col;
	movecursor();
	return;
}

void
cons_putc(int c)
{
	int i;
	unsigned char *p;

	switch(c){
	case '\n':
		cgapos = cgapos/Width;
		cgapos = (cgapos+1)*Width;
		break;

	case '\t':
		i = 8 - ((cgapos/2)&7);
		while(i-->0)
			cons_putc(' ');
		break;

	case '\b':
		if(cgapos >= 2)
			cgapos -= 2;
		cons_putc(' ');
		cgapos -= 2;
		break;

	default:
		cgamem[cgapos++] = c;
		cgamem[cgapos++] = Attr;
		break;
	}
	if(cgapos >= Width*Height){
		bcopy(&cgamem[Width], cgamem, Width*(Height-1));
		p = &cgamem[Width*(Height-1)];
		for(i=0; i<Width/2; i++){
			*p++ = ' ';
			*p++ = Attr;
		}
		cgapos = Width*(Height-1);
	}
	movecursor();
}
