/*
 * Quad arithmetic.
 *
 * This library makes the following assumptions:
 *
 *  - The type long long (aka vlong) exists.
 *
 *  - A quad variable is exactly twice as long as `long'.
 *
 *  - The machine's arithmetic is two's complement.
 *
 * This library can provide 128-bit arithmetic on a machine with 128-bit
 * quads and 64-bit longs, for instance, or 96-bit arithmetic on machines
 * with 48-bit longs.
 */

#include <u.h>

/*
 * Depending on the desired operation, we view a `long long' (aka quad_t) in
 * one or more of the following formats.
 */
union uu {
	vlong	q;		/* as a (signed) quad */
	vlong	uq;		/* as an unsigned quad */
	long	sl[2];		/* as two signed longs */
	ulong	ul[2];		/* as two unsigned longs */
};

/*
 * Define high and low longwords.
 */
#define	H		_QUAD_HIGHWORD
#define	L		_QUAD_LOWWORD

/*
 * Total number of bits in a vlong and in the pieces that make it up.
 * These are used for shifting, and also below for halfword extraction
 * and assembly.
 */
#define CHAR_BIT	8
#define	QUAD_BITS	(sizeof(vlong) * CHAR_BIT)
#define	LONG_BITS	(sizeof(long) * CHAR_BIT)
#define	HALF_BITS	(sizeof(long) * CHAR_BIT / 2)

/*
 * Extract high and low shortwords from longword, and move low shortword of
 * longword to upper half of long, i.e., produce the upper longword of
 * ((vlong)(x) << (number_of_bits_in_long/2)).  (`x' must actually be ulong.)
 *
 * These are used in the multiply code, to split a longword into upper
 * and lower halves, and to reassemble a product as a vlong, shifted left
 * (sizeof(long)*CHAR_BIT/2).
 */
#define	HHALF(x)	((x) >> HALF_BITS)
#define	LHALF(x)	((x) & ((1 << HALF_BITS) - 1))
#define	LHUP(x)		((x) << HALF_BITS)

typedef unsigned int	qshift_t;

vlong		__ashldi3(vlong, qshift_t);
vlong		__ashrdi3(vlong, qshift_t);
vlong		__divdi3(vlong a, vlong b);
vlong		__lshrdi3(vlong, qshift_t);
vlong		__moddi3(vlong a, vlong b);
uvlong		__qdivrem(uvlong u, uvlong v, uvlong *rem);
uvlong		__udivdi3(uvlong a, uvlong b);
uvlong		__umoddi3(uvlong a, uvlong b);
int		__ucmpdi2(uvlong a, uvlong b);
